@@ -9,7 +9,7 @@ from ecmtool import mpi_wrapper
 
 
 def normalize_rows(M):
-    row_max = M.max(axis=1)
+    row_max = abs(M.max(axis=1))
     return M / np.transpose(np.asarray(np.asmatrix(row_max, dtype='object'), dtype='object'))
 
 
@@ -146,18 +146,25 @@ def get_conversion_cone(N, external_metabolites=[], reversible_reactions=[], inp
         # Calculate generating set of the dual of our initial conversion cone C0, C0*
         rays = get_extreme_rays(np.append(linearities, G_rev, axis=0), G_irrev, verbose=verbose)
 
-        if rays.shape[0] == 0:
-            print('Warning: given system has no nonzero inequalities H. Returning empty conversion cone.')
-            return to_fractions(np.ndarray(shape=(0, G.shape[1])))
+        # if rays.shape[0] == 0:
+        #     print('Warning: given system has no nonzero inequalities H. Returning empty conversion cone.')
+        #     return to_fractions(np.ndarray(shape=(0, G.shape[1])))
 
         if verbose:
             print('Deflating H')
-        rays_deflated = deflate_matrix(rays, external_metabolites)
+        if rays.shape[0] == 0:
+            mp_print('Warning: first polco-application did not give any rays. Check if this is expected behaviour.')
+            rays_deflated = rays
+        else:
+            rays_deflated = deflate_matrix(rays, external_metabolites)
 
         if verbose:
             print('Expanding H with metabolite direction constraints')
         # Add bidirectional (in- and output) metabolites in reverse direction
-        rays_split = split_columns(rays_deflated, in_out_indices) if not only_rays else rays_deflated
+        if rays_deflated.shape[0] == 0:
+            rays_split = rays_deflated
+        else:
+            rays_split = split_columns(rays_deflated, in_out_indices) if not only_rays else rays_deflated
         linearities_split = split_columns(linearities_deflated, in_out_indices) if not only_rays else linearities_deflated
 
         H_ineq = rays_split
@@ -208,25 +215,39 @@ def get_conversion_cone(N, external_metabolites=[], reversible_reactions=[], inp
 
         # Remove duplicates from H_ineq and H_eq
         if redund_after_polco:
-            H_ineq = unique(np.transpose(normalize_columns_fraction(np.transpose(H_ineq))))
-            H_eq = unique(np.transpose(normalize_columns_fraction(np.transpose(H_eq))))
+            H_ineq_original = H_ineq
+            H_ineq_normalized = np.transpose(normalize_columns(np.transpose(H_ineq.astype(dtype='float')), verbose=verbose))
+            # unique_inds = find_unique_inds(H_ineq_normalized, verbose=verbose, tol=1e-9)
+            # H_ineq_float = H_ineq_normalized[unique_inds, :]
+            # H_ineq_original = H_ineq_original[unique_inds, :]
+            H_ineq_float, unique_inds = np.unique(H_ineq_normalized, axis=0, return_index=True)
+            H_ineq_original = H_ineq_original[unique_inds,:]
+
+            # H_ineq_float = unique(H_ineq_normalized)
+
+            # Find out if rows have been thrown away, and if so, do that as well
+            # unique_inds = find_remaining_rows(H_ineq_float, H_ineq_normalized, verbose=verbose)
+            # H_ineq_original = H_ineq_original[unique_inds, :]
     else:
+        H_ineq_float = []
         H_ineq = []
         H_eq = []
 
-    H_ineq = mpi_wrapper.world_allgather(H_ineq)
-    H_ineq = H_ineq[0]
-    H_eq = mpi_wrapper.world_allgather(H_eq)
-    H_eq = H_eq[0]
-    print("Size of H_eq after communication step:", H_eq.shape[0], H_eq.shape[1])
-
-    use_custom_redund = True  # If set to false, redundancy removal with redund from lrslib is used
     if redund_after_polco:
+        H_ineq_float = mpi_wrapper.world_allgather(H_ineq_float)
+        H_ineq_float = H_ineq_float[0]
+        H_ineq = mpi_wrapper.world_allgather(H_ineq)
+        H_ineq = H_ineq[0]
+        H_eq = mpi_wrapper.world_allgather(H_eq)
+        H_eq = H_eq[0]
+        if verbose:
+            mp_print("Size of H_eq after communication step:", H_eq.shape[0], H_eq.shape[1])
+
+        use_custom_redund = True  # If set to false, redundancy removal with redund from lrslib is used
         if use_custom_redund:
             mp_print('Using custom redundancy removal')
             t1 = time()
-            H_ineq_transpose, cycle_rays = drop_redundant_rays(np.transpose(H_ineq), rays_are_unique=True, linearities=False, normalised=True)
-            H_ineq = np.transpose(H_ineq_transpose)
+            nonred_inds_ineq, cycle_rays = drop_redundant_rays(np.transpose(H_ineq_float), rays_are_unique=True, linearities=False, normalised=True)
             mp_print("Custom redund took %f sec" % (time()-t1))
 
             t1 = time()
@@ -242,6 +263,9 @@ def get_conversion_cone(N, external_metabolites=[], reversible_reactions=[], inp
             mp_print("Redund took %f sec" % (time() - t2))
 
     if mpi_wrapper.is_first_process():
+        if redund_after_polco:
+            if use_custom_redund:
+                H_ineq = H_ineq_original[nonred_inds_ineq, :]
         print("Size of H_ineq after redund:", H_ineq.shape[0], H_ineq.shape[1])
         print("Size of H_eq after redund:", H_eq.shape[0], H_eq.shape[1])
         count_after_ineq = len(H_ineq)
